import os
import cv2
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image
import glob
import sys
from numba import jit, cuda
from numba import njit, literal_unroll, prange
from numba.typed import List
import math
from vtk_test import numpy_to_vti
import shutil
import time


# This function is used to create two lists, one with all the background pixels in the studied area, the other for
# the foreground pixels.

@jit
def fill(file, tophat, mask, h1, w1):
    background = List()
    foreground = List()
    background.append(120)
    foreground.append(120)

    for i in range(h1):
        for j in range(w1):
            if tophat[i, j] == 0 and mask[i, j] == True:
                background.append(file[i, j])
            elif tophat[i, j] != 0 and mask[i, j] == True:
                foreground.append(file[i, j])

    return background[1:], foreground[1:]


#  This function checks if there is a container in the data set

def testing(img, w, h, test):
    center = [int(w / 2), int(h / 2)]

    if test == 0:
        circles = cv2.HoughCircles(img, cv2.HOUGH_GRADIENT, 1, 1000,
                                   param1=20, param2=30,
                                   minRadius=int(0.95 * (min(center[0], center[1], w - center[0], h - center[1]))),
                                   maxRadius=int((min(center[0], center[1], w - center[0], h - center[1]))))

        if np.all(circles == None):
            print('No circle detected')
            return 0

    print('circle detected')

    return 1


# This function is used to detect containers in the pictures

def hough(img, w, h, flag, rayon, ref, ind, param2, param1, seuil):
    center = [int(w / 2), int(h / 2)]

    if flag == 0:

        circles = cv2.HoughCircles(img, cv2.HOUGH_GRADIENT, 1, 1000,
                                   param1=param3, param2=param4,
                                   minRadius=int(0.75 * (min(center[0], center[1], w - center[0], h - center[1]))),
                                   maxRadius=int((min(center[0], center[1], w - center[0], h - center[1]))))

        if np.all(circles == None):

            print('ok1')

            circles = cv2.HoughCircles(img, cv2.HOUGH_GRADIENT, 1, 1000,
                                       param1=20, param2=30,
                                       minRadius=int(
                                           0.6 * (min(center[0], center[1], w - center[0], h - center[1]))),
                                       maxRadius=int((min(center[0], center[1], w - center[0], h - center[1]))))

            mini = np.min(circles[:, :, 2])

            x_center = circles[0, 0, 0]
            y_center = circles[0, 0, 1]

            ref = mini - 0.01 * mini

            return mini - 0.01 * mini, ref, ind, x_center, y_center


        else:
            x_center = circles[0, 0, 0]
            y_center = circles[0, 0, 1]
            mini = np.min(circles[:, :, 2])
            print('gg', circles[:, :, 2])

            mask = create_circular_mask2(h, w, radius=mini - 0.05 * mini)
            masked_img = img.copy()
            masked_img[~mask] = 0

            circles = cv2.HoughCircles(masked_img, cv2.HOUGH_GRADIENT, 1, 1000,

                                       param1=param1, param2=param2,
                                       minRadius=int(
                                           0.7 * (min(center[0], center[1], w - center[0], h - center[1]))),
                                       maxRadius=int(0.9 * (min(center[0], center[1], w - center[0], h - center[1]))))

            if np.all(circles == None):
                print('1 circle detected')

                print(mini)

                ref = mini - 0.01 * mini

                return mini - 0.01 * mini, ref, ind, x_center, y_center

            else:

                x_center = circles[0, 0, 0]
                y_center = circles[0, 0, 1]

                print('2 circles detected')
                mini = np.min(circles[:, :, 2])

                print(mini)

                if seuil == 1.8:
                    taux = 0.02
                else:
                    taux = 0.035

                ref = mini - taux * mini

                return mini - taux * mini, ref, ind, x_center, y_center


    else:

        if seuil == 1.8:
            dist = 1000

            param2 = 28

            param1 = 48

        elif seuil == 3:
            dist = 1000
            param2 = 19

            param1 = 48

        else:
            dist = 800

            param2 = 12

            param1 = 20

        print('dist', dist)

        circles = cv2.HoughCircles(img, cv2.HOUGH_GRADIENT, 1, dist,
                                   param1=param1, param2=param2,
                                   minRadius=int(rayon - 20),
                                   maxRadius=int(rayon + 20))

        if np.all(circles == None):

            if ind == 0:
                return ref, ref, ind, param2, param1

            else:
                return rayon, rayon, ind, param2, param1

        mini = np.min(circles[:, :, 2])

        print('mini', mini)

        if 0.98 * ref < mini < ref or mini > ref:

            return ref, mini, ind, param2, param1

        else:
            ind = 1
            if seuil == 1.8:
                dist = 490
                param2 = 19

                param1 = 48


            elif seuil == 3:

                param2 = 5
                param1 = 48
                dist = 490


            else:

                param2 = 5
                param1 = 20
                dist = 400

            return mini, mini, ind, param2, param1


# This function crops the picture in a circular way, giving it the appropriate center and radius of the mask.

def create_circular_mask(h, w, center=None, radius=None):
    if center is None:  # use the middle of the image
        center = [int(w / 2), int(h / 2)]

    if radius is None:  # use the smallest distance between the center and image walls
        radius = min(center[0], center[1], w - center[0], h - center[1])

    else:
        radius = radius * min(center[0], center[1], w - center[0], h - center[1])

    Y, X = np.ogrid[:h, :w]
    dist_from_center = np.sqrt((X - center[0]) ** 2 + (Y - center[1]) ** 2)

    if radius == 0:
        mask = np.ones((h, w))

    else:
        if center != [int(w / 2), int(h / 2)]:
            radius = radius + (1 / 3) * radius

        mask = dist_from_center <= radius

    return mask


def create_circular_mask2(h, w, center=None, radius=None):
    if center is None:  # use the middle of the image
        center = [int(w / 2), int(h / 2)]
    if radius is None:  # use the smallest distance between the center and image walls
        radius = min(center[0], center[1], w - center[0], h - center[1])

    Y, X = np.ogrid[:h, :w]
    dist_from_center = np.sqrt((X - center[0]) ** 2 + (Y - center[1]) ** 2)

    mask = dist_from_center <= radius

    return mask


# This function keeps the main information in the pictures.


def cleaning4(image, seuil):
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (2, 2))

    tophat = cv2.morphologyEx(image, cv2.MORPH_TOPHAT, kernel)
    im = tophat

    return tophat


# This function ensure that all the objects touching the border of the mask are erased if they are small enough
# (no loss of important data)

@jit
def cleaning5(edge, h1, w1, labels4, rayon_init):
    center_x = h1 / 2
    center_y = w1 / 2
    del_list = List()
    fin_list = List()

    for i in prange(h1):
        for j in prange(w1):
            if edge[i, j] == 255:
                if labels4[i, j] not in del_list:
                    del_list.append(labels4[i, j])

    for k in del_list:

        flag = 0

        if k not in fin_list:
            coords4 = np.argwhere(labels4 == k)

            for j in prange(len(coords4[:, 0]) - 1):

                if math.sqrt(
                        (coords4[j, 0] - center_x) ** 2 + (coords4[j, 1] - center_y) ** 2) < rate * rayon_init:
                    flag = 1

                    break

            if flag == 0:
                fin_list.append(k)

    return fin_list


@jit
def cleaning6(h1, w1, labels4, file):
    for i in prange(h1):
        for j in prange(w1):
            if labels4[i, j] == 1000:
                file[i, j] = 0

    return file


# This function is searching for the appropriate h value for denoising step. Also it is making groups of data sets
#  which have near standard deviation value.

def find_param(center, files):
    index = int(len(files) / 3)
    files = files[:]

    cpt = 0

    flag3 = 0

    for myFile in files:
        file = cv2.imread(myFile, 0)

        h1, w1 = file.shape[0], file.shape[1]

        mask = create_circular_mask(h1, w1, center=center, radius=1 / 3)
        masked_img = file.copy()
        masked_img[~mask] = 0

        kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (2, 2))
        tophat = cv2.morphologyEx(masked_img, cv2.MORPH_TOPHAT, kernel)
        blackhat = cv2.morphologyEx(masked_img, cv2.MORPH_BLACKHAT, kernel)
        im = blackhat + tophat

        counts = np.bincount(file.ravel()).argmax()

        val = np.count_nonzero(file == counts)

        ratio = val / len(file.ravel())

        if flag3 == 0:
            print('value', np.std(file))

            if np.std(file) > 7:
                seuil = 5

            elif np.std(file) > 5:
                seuil = 3


            elif np.std(file) > 1.7:
                seuil = 1.8

        flag3 = 1

        ret, tophat = cv2.threshold(im, seuil, 255, cv2.THRESH_BINARY)

        background, foreground = fill(file, tophat, mask, h1, w1)

        print('% background', float(len(background) / (len(background) + len(foreground))))
        print('% foreground', float(len(foreground) / (len(background) + len(foreground))))

        if float(len(background) / (len(background) + len(foreground))) > 0.1 and float(
                len(foreground) / (len(background) + len(foreground))) > 0.2:
            sigma = np.std(background)

            h = 1.2 * sigma

            return h, seuil

        cpt += 1

    return 0, 0


Datas = ['Archie_04', 'Probe_7', 'Z275', 'C7', 'Pseudoskorpion', 'Eucrib_03', 'Gama_01', 'Screwjoint']

for data in Datas:

    print(data)

    files = glob.glob('/data/' + str(data) + '/slices_8bit/*.tif')
    files.sort(key=lambda x: int(''.join(filter(str.isdigit, x))))

    X_data = []

    image = Image.open(files[0]).convert('L')
    image = np.array(image)

    h1, w1 = image.shape[0], image.shape[1]

    # Looking for different part of the pictures to find the best h value.

    for n in [[int(w1 / 2), int(h1 / 2)], [int(w1 / 2), int(h1 / 3)], [int(w1 / 3), int(h1 / 2)],
              [int(2 * w1 / 3), int(h1 / 2)],
              [int(w1 / 2), int(2 * h1 / 3)]]:
        h, seuil = find_param(n, files)

        print('seuil', seuil)
        if h != 0:
            print('h', h)
            break

    cpt2 = 0
    radius = 0
    flag = 0
    rayon = 0
    test = 0
    cpt8 = 1

    ind = 0

    rate = 0.99

    # defining hough parameters.

    if seuil == 1.8:

        param2 = 15

        param1 = 30

        param3 = 20
        param4 = 39

    else:

        param2 = 30
        param1 = 30

        param3 = 20
        param4 = 39

    fact = h1 / 256
    value = round(fact)
    compteur = 1

    for myFile in files:
        print('param1', param1)
        print('param2', param2)

        file = cv2.imread(myFile, 0)

        h1, w1 = file.shape[0], file.shape[1]

        file = cv2.fastNlMeansDenoising(file, h=h)

        print('num', cpt2)

        cpt2 += 1

        if test == 0:
            rayon_init = testing(file, w1, h1, test)

        test = 1

        kernel2 = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (3, 3))
        file2 = cv2.dilate(file, kernel2, iterations=4)
        file2 = cv2.morphologyEx(file2, cv2.MORPH_CLOSE, kernel2, iterations=4)

        if rayon_init != 0:

            if flag == 0:
                rayon_init, ref, ind, x_center, y_center = hough(file, w1, h1, flag, 0, 0, ind, param2, param1, seuil)
                last = rayon_init

                flag = 1

            rayon_init, last, ind, param2, param1 = hough(file, w1, h1, flag, last, ref, ind, param2,
                                                          param1, seuil)

            if ind == 1:
                if seuil == 1.8:
                    rate = 0.7

                elif seuil == 3:
                    rate = 0.8

                else:
                    rate = 0.8

            print('rate', rate)

            print('ref', ref)

            file = cleaning4(file, seuil)

            print('rad', rayon_init)

            file3 = cv2.dilate(file, kernel2, iterations=2)
            file3 = cv2.morphologyEx(file3, cv2.MORPH_CLOSE, kernel2, iterations=2)
            ret, file3 = cv2.threshold(file3, seuil, 255, cv2.THRESH_BINARY)

            ret, labels4 = cv2.connectedComponents(file3)

            mask = create_circular_mask2(h1, w1, center=[x_center, y_center], radius=rayon_init)

            del_list = List()
            edge = cv2.Canny(mask.astype(np.uint8), 0, 1)

            start = time.time()

            fin_list = cleaning5(edge, h1, w1, labels4, rayon_init)

            end = time.time()
            print((str(round(end - start, 4))))

            cpt8 += 1

            for i in fin_list:
                labels4[labels4 == i] = 1000

            file = cleaning6(h1, w1, labels4, file)

            masked_img = file.copy()
            masked_img[~mask] = 0

            full = file.copy()

            kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (7, 7))

            # Reconstruction of possible important pixels erased due to the proximity with container.
            # Only for sensible data set.

            if seuil == 1.8:
                for i in range(50):
                    masked_img = cv2.dilate(masked_img, kernel, iterations=1)
                    masked_img = np.minimum(masked_img, full)

            file = masked_img


        else:
            file = cleaning4(file, seuil)

        file = Image.fromarray(file)

        # Pictures are resized to avoid having a file too big to be displayed on the web page.

        new_width = 256
        new_height = 256
        im = file.resize((new_width, new_height), Image.ANTIALIAS)

        if compteur == value:

            X_data.append(file)

            compteur = 1

        else:

            compteur += 1

    X_data = np.array(X_data)

    # Creation of VTI file

    numpy_to_vti(X_data, (0, 0, 0), (float(X_data.shape[1] / X_data.shape[0]), 1, 1),
                 '/data/vtis/' + str(data) + '_cpu.vti')
